/*----------------------------------------------------------------------
  File    : JNICoCo.java
  Contents: Java interface to CoCoNAD implementation in C
  Author  : Christian Borgelt
  History : 2013.10.21 file created
            2013.11.25 first full version completed
            2013.12.04 pattern spectrum generation function added
            2014.03.06 pattern spectrum estimation added (estpsp())
            2014.05.09 surrogate data via block permutations added
            2014.05.12 parameter border added to coconad function
            2014.07.16 definitions of SEQUENCE and SEQ added
            2014.07.25 spectrum estimation for item sequences added
            2014.10.23 changed from LGPL license to MIT license
            2015.03.05 handling of user abort request improved
            2015.08.13 functions patred() added (for consistency)
            2016.03.24 perfect extension flag replaced by mode
            2016.10.03 adapted to extensions of coconad implementation
            2016.10.23 closed/maximal repository coded in operation mode
            2016.11.04 adapted to fim modules (Pattern, PatternSet etc.)
            2019.10.11 bug in class specification fixed (PatSpecElem)
            2023.07.30 constructor added
----------------------------------------------------------------------*/
package coconad;

import java.util.Random;

import fim.Pattern;
import fim.PatternSet;
import fim.PatSpecElem;

/*--------------------------------------------------------------------*/
/** Class for Java interface to CoCoNAD implementation in C
 *  @author Christian Borgelt
 *  @since  2013.11.21 */
/*--------------------------------------------------------------------*/
public class JNICoCo {

  /*------------------------------------------------------------------*/
  /* Constants                                                        */
  /*------------------------------------------------------------------*/
  /** the version string */
  public static final String VERSION = "1.29 (2019.10.11)";

  /* --- target types --- */
  /** target pattern subtype mask */
  public static final int SUBMASK     = 0x00ff;
  /** target pattern type mask */
  public static final int TYPEMASK    = ~SUBMASK;
  /** target pattern type: item sets */
  public static final int ITEMSET     = 0x0000;
  /** target pattern type: partial permutations (without repetitions) */
  public static final int PERMUTATION = 0x0100;
  /** target pattern type: sequences (with and without repetitions) */
  public static final int SEQUENCE    = 0x0400;
  /** target pattern subtype: all frequent patterns */
  public static final int FREQUENT    = 0x0000;
  /** target pattern subtype: all frequent patterns */
  public static final int ALL         = 0x0000;
  /** target pattern subtype: all closed (frequent) patterns */
  public static final int CLOSED      = 0x0001;
  /** target pattern subtype: all maximal (frequent) patterns */
  public static final int MAXIMAL     = 0x0002;
  /** target pattern subtype flag: closed/maximal w.r.t. appending */
  public static final int APPEND      = 0x0004;

  /* --- CoCoNAD algorithm variants --- */
  /** algorithm variant: automatic choice based on target type */
  public static final int AUTO        = 0;
  /** algorithm variant: basic version (low memory consumption, slow) */
  public static final int BASIC       = 1;
  /** algorithm variant: refined version (spike and train filtering) */
  public static final int FILTER      = 2;
  /** algorithm variant: refined version (closed/maximal repository) */
  public static final int CMREPO      = 3;
  /** algorithm variant: refined version (closed/maximal repository) */
  public static final int REPO        = 3;

  /* --- CoCoNAD operation mode flags --- */
  /** processing mode: no special flags */
  public static final int NONE        = 0x0000;
  /** processing mode: do not use perfect extension pruning */
  public static final int NOPERFECT   = 0x0002; /* x */
  /** processing mode: do not reorder trains by their length */
  public static final int NOSORT      = 0x0004; /* p */
  /** processing mode: repository is a chain of prefix trees */
  public static final int REPOCHAIN   = 0x0000;
  /** processing mode: repository is a single prefix trees */
  public static final int REPOTREE    = 0x0100;
  /** processing mode: repository is a list/an array */
  public static final int REPOLIST    = 0x0200;

  /* --- surrogate data generation methods --- */
  /** surrogate method: identity (keep original data) */
  public static final int IDENTITY    = 0;
  /** surrogate method: point/spike time rendomization */
  public static final int RANDOM      = 1;
  /** surrogate method: point/spike dithering/displacement */
  public static final int DITHER      = 2;
  /** surrogate method: train shifting/dithering */
  public static final int SHIFT       = 3;
  /** surrogate method: sampling from kernel estimate */
  public static final int KERNEST     = 4;
  /** surrogate method: dithered point/spike permutation */
  public static final int PERMUTE     = 5;
  /** surrogate method: dithered blocked permutation */
  public static final int BLOCK       = 6;

  /* --- random number density functions --- */
  /** random number density function: uniform
   *  (identical to <code>RECT</code> */
  public static final int UNIFORM     = 0;
  /** random number density function: rectangular
   *  (identical to <code>UNIFORM</code> */
  public static final int RECT        = 0;
  /** random number density function: triangular */
  public static final int TRIANG      = 1;
  /** random number density function: Gaussian
   *  (identical to <code>NORMAL</code> */
  public static final int GAUSS       = 2;
  /** random number density function: normal
   *  (identical to <code>GAUSS</code> */
  public static final int NORMAL      = 2;

  /* --- pattern spectrum report formats --- */
  /** pattern spectrum report format: objects */
  public static final String OBJECTS  = "-";
  /** pattern spectrum report format: columns */
  public static final String COLUMNS  = "|";

  /* --- pattern set reduction methods --- */
  /** pattern set reduction: none (keep all patterns after filtering) */
  public static final int KEEP        = PatternSet.KEEP;
  /** pattern set reduction: excess coincidences (zb,cb-ca) */
  public static final int COINS0      = PatternSet.COINS0;
  /** pattern set reduction: excess coincidences (zb,cb-ca+1) */
  public static final int COINS1      = PatternSet.COINS1;
  /** pattern set reduction: excess items/neurons (za-zb+2,ca) */
  public static final int ITEMS2      = PatternSet.ITEMS2;
  /** pattern set reduction: covered points/spikes za*ca : zb*cb */
  public static final int COVER0      = PatternSet.COVER0;
  /** pattern set reduction: covered points/spikes
   *                         (za-1)*ca : (zb-1)*cb */
  public static final int COVER1      = PatternSet.COVER1;
  /** pattern set reduction: combined lenient
   *                         (z, break rejection tie) */
  public static final int LENIENT0    = PatternSet.LENIENT0;
  /** pattern set reduction: combined lenient
   *                         (z-1, break rejection tie) */
  public static final int LENIENT1    = PatternSet.LENIENT1;
  /** pattern set reduction: combined strict
   *                         (z, force decision) */
  public static final int STRICT0     = PatternSet.STRICT0;
  /** pattern set reduction: combined strict
   *                         (z-1, force decision) */
  public static final int STRICT1     = PatternSet.STRICT1;

  /*------------------------------------------------------------------*/
  /** Constructor.
   *  @since  2023.07.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public JNICoCo ()
  { }

  /*------------------------------------------------------------------*/
  /** Java interface to CoCoNAD implementation in C.
   *  @param  trains array of (spike) trains to analyze, each of
   *                 which is an array of floating point numbers
   *  @param  target type of the patterns to find
   *                 (<code>ITEMSET</code>, <code>PERMUTATION</code>
   *                  <code>PERMUT</code>, <code>SEQUENCE</code> or
   *                  <code>SEQ</code> as the target pattern type and
   *                  <code>FREQUENT</code>, <code>ALL</code>,
   *                  <code>CLOSED</code> or <code>MAXIMAL</code>
   *                  as the target pattern subtype, with possibly
   *                  added flag <code>APPEND</code>)
   *  @param  width  width of the time window/maximum time span
   *  @param  supp   minimum support of an item set
   *  @param  zmin   minimum number of items per item set
   *  @param  zmax   maximum number of items per item set
   *  @param  report values to report with an item set<br>
   *                 "a": item set support (number of coincidences)<br>
   *                 "|": pattern spectrum in column format<br>
   *                 "-": pattern spectrum with <code>PatSpecElem</code><br>
   *                 "=": pattern spectrum with <code>PatSpecElem</code>
   *  @param  algo   algorithm variant to use
   *                 (<code>BASIC</code>, <code>FILTER</code> or
   *                  <code>CMREPO</code>)
   *  @param  mode   processing mode to use (<code>NONE</code> or flags
   *                 <code>NOPERFECT</code> or <code>NOSORT</code>)
   *  @param  border array of support thresholds per item set size
   *  @return if report = "a": an array with two elements, the first
   *          of which is an array that contains the item sets as
   *          integer arrays, while the second is an integer array
   *          containing the support values (each corresponding to the
   *          item set at the same array index in the first array).<br>
   *          if report = "|": an array with three elements; the first
   *          array contains the item set sizes as integers, the second
   *          array contains the support values as integers (that is,
   *          corresponding elements of the first and the second array
   *          form a pattern signature), and the third array contains
   *          the frequency (number of occurrences) of the size/support
   *          pair (at the same array index in the first two arrays)
   *          as a double precision floating point value.<br>
   *          if report = '=' or '-': an array with objects of type
   *          <code>PatSpecElem</code>, each of which specifies a
   *          pattern signature together with its occurrence frequency.
   *  @since  2013.11.21 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static native Object[] coconad (double[][] trains,
    int target, double width, int supp, int zmin, int zmax,
    String report, int algo, int mode, int[] border);

  /*------------------------------------------------------------------*/
  /** Java interface to CoCoNAD implementation in C
   *  (wrapper with Java objects).
   *  @param  trains (spike) trains to analyze
   *  @param  target type of the patterns to find
   *                 (<code>ITEMSET</code>, <code>PERMUTATION</code>
   *                  <code>PERMUT</code>, <code>SEQUENCE</code> or
   *                  <code>SEQ</code> as the target pattern type and
   *                  <code>FREQUENT</code>, <code>ALL</code>,
   *                  <code>CLOSED</code> or <code>MAXIMAL</code>
   *                  as the target pattern subtype, with possibly
   *                  added flag <code>APPEND</code>)
   *  @param  width  width of the time window/maximum time span
   *  @param  supp   minimum support of an item set
   *  @param  zmin   minimum number of items per item set
   *  @param  zmax   maximum number of items per item set
   *  @param  algo   algorithm variant to use
   *                 (<code>BASIC</code>, <code>FILTER</code> or
   *                  <code>CMREPO</code>)
   *  @param  mode   processing mode to use (<code>NONE</code> or flags
   *                 <code>NOPERFECT</code> or <code>NOSORT</code>)
   *  @param  border array of support thresholds per item set size
   *  @return a set of (frequent) patterns
   *  @since  2014.10.23 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static PatternSet coconad (TrainSet trains,
    int target, double width, int supp, int zmin, int zmax,
    int algo, int mode, int[] border)
  {                             /* --- wrapper with Java types */
    return new PatternSet(trains.getItemBase(),
                          coconad(trains.getAllTrains(), target,
                                  width, supp, zmin, zmax, "a",
                                  algo, mode, border));
  }  /* coconad() */

  /*------------------------------------------------------------------*/
  /** Pattern spectrum generation with surrogate data sets.
   *  @param  trains array of (spike) trains to analyze, each of
   *                 which is an array of floating point numbers
   *  @param  target type of the patterns to find
   *                 (<code>ITEMSET</code>, <code>PERMUTATION</code>
   *                  <code>PERMUT</code>, <code>SEQUENCE</code> or
   *                  <code>SEQ</code> as the target pattern type and
   *                  <code>FREQUENT</code>, <code>ALL</code>,
   *                  <code>CLOSED</code> or <code>MAXIMAL</code>
   *                  as the target pattern subtype, with possibly
   *                  added flag <code>APPEND</code>)
   *  @param  width  width of the time window/maximum time span
   *  @param  supp   minimum support of an item set
   *  @param  zmin   minimum number of items per item set
   *  @param  zmax   maximum number of items per item set
   *  @param  report format in which to report the pattern spectrum<br>
   *                 "|": pattern spectrum in column format<br>
   *                 "-": pattern spectrum with <code>PatSpecElem</code><br>
   *                 "=": pattern spectrum with <code>PatSpecElem</code>
   *  @param  algo   algorithm variant to use
   *                 (<code>BASIC</code>, <code>FILTER</code> or
   *                  <code>CMREPO</code>)
   *  @param  mode   processing mode to use (<code>NONE</code> or flags
   *                 <code>NOPERFECT</code> or <code>NOSORT</code>)
   *  @param  cnt    number of surrogate data sets to generate
   *  @param  beg    beginning of allowed span of points
   *  @param  end    end       of allowed span of points
   *  @param  surr   surrogate data generation method
   *                 (<code>IDENTITY</code>, <code>RANDOM</code>,
   *                  <code>DITHER</code>, <code>SHIFT</code>,
   *                  <code>KERNEST</code> or <code>PERMUTE</code>)
   *  @param  rand   random number density function
   *                 (for point/spike displacements)
   *                 (<code>UNIFORM</code>, <code>RECT</code>,
   *                  <code>TRIANG</code>, <code>GAUSS</code> or
   *                  <code>NORMAL</code>)
   *  @param  sigma  standard deviation or half the base width
   *                 of the rectangular/triangular density function
   *  @param  delta  block size for blocked permutations
   *  @param  seed   seed value for random number generator
   *  @param  cpus   number of cpus/threads to use
   *  @param  ctrl   control array (progress indicator, stop flag)
   *  @return if report = "|": an array with three elements; the first
   *          array contains the item set sizes as integers, the second
   *          array contains the support values as integers (that is,
   *          corresponding elements of the first and the second array
   *          form a pattern signature), and the third array contains
   *          the frequency (number of occurrences) of the size/support
   *          pair (at the same array index in the first two arrays)
   *          as a double precision floating point value.<br>
   *          if report = '=' or '-': an array with objects of type
   *          <code>PatSpecElem</code>, each of which specifies a
   *          pattern signature together with its occurrence frequency.
   *  @since  2013.12.04 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static native Object[] genpsp (double[][] trains,
    int target, double width, int supp, int zmin, int zmax,
    String report, int algo, int mode, int cnt,
    double beg, double end, int surr, int rand, double sigma,
    double delta, int seed, int cpus, int[] ctrl);

  /*------------------------------------------------------------------*/
  /** Pattern spectrum generation with surrogate data sets
   *  (wrapper with java objects).
   *  @param  trains (spike) trains to analyze
   *  @param  target type of the patterns to find
   *                 (<code>ITEMSET</code>, <code>PERMUTATION</code>
   *                  <code>PERMUT</code>, <code>SEQUENCE</code> or
   *                  <code>SEQ</code> as the target pattern type and
   *                  <code>FREQUENT</code>, <code>ALL</code>,
   *                  <code>CLOSED</code> or <code>MAXIMAL</code>
   *                  as the target pattern subtype, with possibly
   *                  added flag <code>APPEND</code>)
   *  @param  width  width of the time window/maximum time span
   *  @param  supp   minimum support of an item set
   *  @param  zmin   minimum number of items per item set
   *  @param  zmax   maximum number of items per item set
   *  @param  algo   algorithm variant to use
   *                 (<code>BASIC</code>, <code>FILTER</code> or
   *                  <code>CMREPO</code>)
   *  @param  mode   processing mode to use (<code>NONE</code> or flags
   *                 <code>NOPERFECT</code> or <code>NOSORT</code>)
   *  @param  cnt    number of surrogate data sets to generate
   *  @param  beg    beginning of allowed span of points
   *  @param  end    end       of allowed span of points
   *  @param  surr   surrogate data generation method
   *                 (<code>IDENTITY</code>, <code>RANDOM</code>,
   *                  <code>DITHER</code>, <code>SHIFT</code>,
   *                  <code>KERNEST</code> or <code>PERMUTE</code>)
   *  @param  rand   random number density function
   *                 (for point/spike displacements)
   *                 (<code>UNIFORM</code>, <code>RECT</code>,
   *                  <code>TRIANG</code>, <code>GAUSS</code> or
   *                  <code>NORMAL</code>)
   *  @param  sigma  standard deviation or half the base width
   *                 of the rectangular/triangular density function
   *  @param  delta  block size for blocked permutations
   *  @param  seed   seed value for random number generator
   *  @param  cpus   number of cpus/threads to use
   *  @param  ctrl   control array (progress indicator, stop flag)
   *  @return an array with objects of type <code>PatSpecElem</code>,
   *          each of which specifies a pattern signature together
   *          with its occurrence frequency
   *  @since  2014.10.23 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static PatSpecElem[] genpsp (TrainSet trains,
    int target, double width, int supp, int zmin, int zmax,
    int algo, int mode, int cnt, double beg, double end,
    int surr, int rand, double sigma, double delta, int seed,
    int cpus, int[] ctrl)
  {                             /* --- wrapper with Java types */
    return (PatSpecElem[])genpsp(trains.getAllTrains(), target,
                                 width, supp, zmin, zmax, "=",
                                 algo, mode, cnt, beg, end,
                                 surr, rand, sigma, delta, seed,
                                 cpus, ctrl);
  }  /* genpsp() */

  /*------------------------------------------------------------------*/
  /** Estimate a pattern spectrum from data characteristics.
   *  @param  trains array of (spike) trains to analyze, each of
   *                 which is an array of floating point numbers
   *  @param  target type of the patterns to find
   *                 (<code>ITEMSET</code>, <code>PERMUTATION</code>
   *                  <code>PERMUT</code>, <code>SEQUENCE</code> or
   *                  <code>SEQ</code> as the target pattern type and
   *                  <code>FREQUENT</code>, <code>ALL</code>,
   *                  <code>CLOSED</code> or <code>MAXIMAL</code>
   *                  as the target pattern subtype)
   *  @param  width  width of the time window/maximum time span
   *  @param  supp   minimum support of an item set
   *  @param  zmin   minimum number of items per item set
   *  @param  zmax   maximum number of items per item set
   *  @param  report format in which to report the pattern spectrum<br>
   *                 "|": pattern spectrum in column format<br>
   *                 "-": pattern spectrum with <code>PatSpecElem</code><br>
   *                 "=": pattern spectrum with <code>PatSpecElem</code>
   *  @param  equiv  equivalent number of surrogate data sets
   *  @param  alpha  probability dispersion factor
   *  @param  smpls  number of samples per item set size
   *  @param  seed   seed value for random number generator
   *  @return if report = "|": an array with three elements; the first
   *          array contains the item set sizes as integers, the second
   *          array contains the support values as integers (that is,
   *          corresponding elements of the first and the second array
   *          form a pattern signature), and the third array contains
   *          the frequency (number of occurrences) of the size/support
   *          pair (at the same array index in the first two arrays)
   *          as a double precision floating point value.<br>
   *          if report = '=' or '-': an array with objects of type
   *          <code>PatSpecElem</code>, each of which specifies a
   *          pattern signature together with its occurrence frequency.
   *  @since  2014.03.06 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static native Object[] estpsp (double[][] trains,
    int target, double width, int supp, int zmin, int zmax,
    String report, int equiv, double alpha, int smpls, int seed);

  /*------------------------------------------------------------------*/
  /** Estimate a pattern spectrum from data characteristics
   *  (wrapper with Java objects).
   *  @param  trains (spike) trains to analyze
   *  @param  target type of the patterns to find
   *                 (<code>ITEMSET</code>, <code>PERMUTATION</code>
   *                  <code>PERMUT</code>, <code>SEQUENCE</code> or
   *                  <code>SEQ</code> as the target pattern type and
   *                  <code>FREQUENT</code>, <code>ALL</code>,
   *                  <code>CLOSED</code> or <code>MAXIMAL</code>
   *                  as the target pattern subtype)
   *  @param  width  width of the time window/maximum time span
   *  @param  supp   minimum support of an item set
   *  @param  zmin   minimum number of items per item set
   *  @param  zmax   maximum number of items per item set
   *  @param  equiv  equivalent number of surrogate data sets
   *  @param  alpha  probability dispersion factor
   *  @param  smpls  number of samples per item set size
   *  @param  seed   seed value for random number generator
   *  @return an array with objects of type <code>PatSpecElem</code>,
   *          each of which specifies a pattern signature together
   *          with its occurrence frequency
   *  @since  2014.10.23 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static PatSpecElem[] estpsp (TrainSet trains,
    int target, double width, int supp, int zmin, int zmax,
    int equiv, double alpha, int smpls, int seed)
  {                             /* --- wrapper with Java types */
    return (PatSpecElem[])estpsp(trains.getAllTrains(), target,
                                 width, supp, zmin, zmax, "=",
                                 equiv, alpha, smpls, seed);
  }  /* estpsp() */

  /*------------------------------------------------------------------*/
  /** Extract decision border from pattern spectrum.
   *  @param  psp  the pattern spectrum to process
   *  @return an array of minimum support values per pattern size;
   *          to be indexed by the pattern size
   *  @since  2015.08.13 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static int[] psp2bdr (Object[] psp)
  {                             /* --- extract a decision border */
    int   i;                    /* loop variable */
    int[] sizes;                /* pattern sizes */
    int[] supps;                /* pattern support values */
    int   zmax = -1;            /* maximum pattern size */
    int[] bdr;                  /* created decision border */

    sizes = (int[])psp[0];      /* get the pattern sizes */
    supps = (int[])psp[1];      /* and the support values */
    for (i = 0; i < sizes.length; i++)
      if (sizes[i] > zmax) zmax = sizes[i];
    if (zmax < 0) return new int[0];
    bdr = new int[zmax+1];      /* create the decision border */
    for (i = 0; i < sizes.length; i++)
      if (supps[i] >= bdr[sizes[i]])  /* find largest support */
        bdr[sizes[i]] = supps[i]+1;   /* per pattern size */
    for (i = zmax; --i >= 0; )  /* ensure a monotone border */
      if (bdr[i+1] > bdr[i]) bdr[i] = bdr[i+1];
    bdr[0] = bdr[1] = Integer.MAX_VALUE;
    return bdr;                 /* return the created decision border */
  }  /* psp2bdr() */

  /*------------------------------------------------------------------*/
  /** Extract decision border from pattern spectrum.
   *  @param  psp  the pattern spectrum to process
   *  @return an array of minimum support values per pattern size;
   *          to be indexed by the pattern size
   *  @since  2015.08.13 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static int[] psp2bdr (PatSpecElem[] psp)
  {                             /* --- extract a decision border */
    int   i;                    /* loop variable */
    int   zmax = -1;            /* maximum pattern size */
    int[] bdr;                  /* created decision border */

    for (i = 0; i < psp.length; i++)
      if (psp[i].size > zmax) zmax = psp[i].size;
    if (zmax < 0) return new int[0];
    bdr = new int[zmax+1];      /* create the decision border */
    for (i = 0; i < psp.length; i++)
      if (psp[i].supp >= bdr[psp[i].size])  /* find largest support */
        bdr[psp[i].size] = psp[i].supp+1;   /* per pattern size */
    for (i = zmax; --i >= 0; )  /* ensure a monotone border */
      if (bdr[i+1] > bdr[i]) bdr[i] = bdr[i+1];
    bdr[0] = bdr[1] = Integer.MAX_VALUE;
    return bdr;                 /* return the created decision border */
  }  /* psp2bdr() */

  /*------------------------------------------------------------------*/
  /** Reduce a pattern set.
   *  @param  pats    the pattern set to reduce as an array with two
   *                  elements, the first of which is an array that
   *                  contains the item sets as integer arrays, while
   *                  the second is an integer array containing the
   *                  support values (each corresponding to the item
   *                  set the same array index in the first array)
   *  @param  method  the pattern set reduction method (one of
   *                  <code>KEEP</code>, <code>COINS0</code>,
   *                  <code>COINS1</code>, <code>ITEMS2</code>,
   *                  <code>COVER0</code>, <code>COVER1</code>,
   *                  <code>LENIENT0</code>, <code>LENIENT1</code>,
   *                  <code>STRICT0</code>, <code>STRICT1</code>)
   *  @param  border  the decision border for rejecting patterns
   *  @param  addis   whether to add pattern intersections
   *  @return the reduced pattern set as an array with two elements,
   *          the first of which is an array that contains the item sets
   *          as integer arrays, while the second is an integer array
   *          containing the support values (each corresponding to the
   *          item set the same array index in the first array)
   *  @since  2015.08.13 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static Object[] patred (Object[] pats,
    int method, int[] border, boolean addis)
  {                             /* --- extract a decision border */
    PatternSet pset;            /* patterns as a pattern set */
    Pattern    p;               /* to traverse the patterns */
    int        n;               /* reduced number of patterns */
    int[][]    sets;            /* item sets of the patterns */
    int[]      supps;           /* support values of the patterns */

    pset = new PatternSet(null, pats);
    n    = pset.reduce(method, border, addis);
    if (n == ((int[][])pats[0]).length) return pats;
    sets  = new int[n][];       /* reduce the pattern set and */
    supps = new int[n];         /* create new set and support arrays */
    while (--n >= 0) {          /* traverse the reduced patterns */
      p = pset.getPattern(n);   /* get the next pattern */
      sets [n] = p.getAllItems();
      supps[n] = p.getSupp();   /* store items and support */
    }                           /* in corresp. array elements */
    return new Object[] { sets, supps };
  }  /* patred() */             /* return the reduced pattern set */

  /*------------------------------------------------------------------*/
  /** Reduce a pattern set.
   *  @param  pats    the pattern set to reduce
   *  @param  method  the pattern set reduction method (one of
   *                  <code>KEEP</code>, <code>COINS0</code>,
   *                  <code>COINS1</code>, <code>ITEMS2</code>,
   *                  <code>COVER0</code>, <code>COVER1</code>,
   *                  <code>LENIENT0</code>, <code>LENIENT1</code>,
   *                  <code>STRICT0</code>, <code>STRICT1</code>)
   *  @param  border  the decision border for rejecting patterns
   *  @param  addis   whether to add pattern intersections
   *  @return the reduced pattern set
   *  @since  2015.08.13 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static PatternSet patred (PatternSet pats,
    int method, int[] border, boolean addis)
  {                             /* --- extract a decision border */
    pats.reduce(method, border, addis);
    return pats;                /* reduce the pattern set */
  }  /* patred() */

  /*------------------------------------------------------------------*/
  /** Set the abort state (abort computations or clear abort state).
   *  @param  state  abort state to set (0: clear; != 0: signal abort)
   *  @since  2015.03.05 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static native void abort (int state);

  /*------------------------------------------------------------------*/
  /** Main program for testing.
   *  @param  args the command line arguments
   *  @since  2013.11.21 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static void main (String[] args)
  {                             /* --- main function for testing */
    int         i, k;           /* loop variables */
    int         cnt;            /* number of surrogate data sets */
    double[][]  trains;         /* parallel spike trains */
    Object[]    res;            /* result of coconad call */
    int[]       sizes, supps;   /* sizes and support values */
    double[]    frqs;           /* occurrence frequencies */
    int[][]     pats;           /* found patterns as item id arrays */
    PatSpecElem pse;            /* pattern spectrum element */
    Random      rand;           /* random number generator */
    int[]       ctrl = {0,0};   /* control array (progress indicator) */

    trains = new double[100][]; /* create a spike trains array */
    rand   = new Random();      /* and a random number generator */
    for (i = trains.length; --i >= 0; ) {
      trains[i] = new double[60];
      for (k = trains[i].length; --k >= 0; )
        trains[i][k] = 3.0 *rand.nextDouble();
    }                           /* generate random spike trains */
    if (args.length <= 0) cnt = 1;
    else try { cnt = Integer.parseInt(args[0]); }
    catch (NumberFormatException e) { cnt = 0; }
    if      (cnt <= -4)         /* test pattern spectrum estimation */
      res = estpsp (trains, FREQUENT|ITEMSET,   0.003, 2, 2, -1, "|",
                    10000, 0.5, 1000, 0);
    else if (cnt <= -3)         /* test pattern spectrum estimation */
      res = estpsp (trains, FREQUENT|ITEMSET,   0.003, 2, 2, -1, "=",
                    10000, 0.5, 1000, 0);
    else if (cnt <= -2)         /* test pattern spectrum */
      res = coconad(trains, CLOSED|ITEMSET, 0.003, 2, 2, -1, "|",
                    CMREPO, NONE, null);
    else if (cnt <= -1)         /* test pattern spectrum */
      res = coconad(trains, CLOSED|ITEMSET, 0.003, 2, 2, -1, "=",
                    CMREPO, NONE, null);
    else if (cnt <=  0)         /* test full pattern set */
      res = coconad(trains, CLOSED|ITEMSET, 0.003, 2, 2, -1, "a",
                    CMREPO, NONE, null);
    else                        /* test surrogate generation */
      res = genpsp (trains, CLOSED|ITEMSET, 0.003, 2, 2, -1, "|",
                    CMREPO, NONE, cnt,
                    Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY,
                    PERMUTE, RECT, 0.005, 0.03, 0, 0, ctrl);
    if (cnt == 0) {             /* if full pattern set */
      pats  = (int[][]) res[0]; /* find frequent item sets */
      supps = (int[])   res[1]; /* and get the result parts */
      for (i = 0; i < pats.length; i++) {
        for (k = 0; k < pats[i].length; k++)
          System.out.print(pats[i][k]+" ");
        System.out.println(" ("+supps[i]+")");
      } }                       /* traverse and print item sets */
    else if ((cnt == -1)        /* if pattern spectrum with objects */
    ||       (cnt == -3)) {     /* (PatSpecElem) */
      for (i = 0; i < res.length; i++) {
        pse = (PatSpecElem)res[i];
        System.out.println(pse.size+" "+pse.supp+" : "+pse.freq);
      } }                       /* traverse and print triplets */
    else {                      /* if pattern spectrum with columns */
      sizes = (int[])   res[0]; /* generate a pattern spectrum */
      supps = (int[])   res[1]; /* and get its three columns, */
      frqs  = (double[])res[2]; /* then print the triplets */
      for (i = 0; i < sizes.length; i++)
        System.out.println(sizes[i]+" "+supps[i]+" : "+frqs[i]);
    }                           /* print triplets of pattern spectrum */
  }  /* main() */

  /*------------------------------------------------------------------*/

  static {                      /* --- load the native library */
    System.loadLibrary("JNICoCo");
  }                             /* java -Djava.library.path=... */

}  /* class JNICoCo */
