/*----------------------------------------------------------------------
  File    : Train.java
  Contents: train management for visualization
  Author  : Christian Borgelt
  History : 2013.11.28 file created from ARule.java
            2013.12.13 numeric item id added for train sorting
            2015.08.17 generally renamed point -> time
            2016.04.07 StringBuffer replaced by StringBuilder
            2016.08.25 bug in fuction filter fixed (|| instead of &&)
----------------------------------------------------------------------*/
package coconad;

import java.util.Arrays;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.io.Serializable;

import util.IdMap;
import util.TableReader;
import util.TableWriter;

/*--------------------------------------------------------------------*/
/** Class for trains.
 *  @author Christian Borgelt
 *  @since  2013.11.28 */
/*--------------------------------------------------------------------*/
public class Train implements Serializable {

  /*------------------------------------------------------------------*/
  /*  constants                                                       */
  /*------------------------------------------------------------------*/
  private static final long serialVersionUID = 0x00010001L;
  /** the block size for the times array */
  private static final int BLKSIZE   = 32;
  /** read/write mode: train with preceding item */
  public static final int ITEM_TRAIN = 0;
  /** read/write mode: pure train, that is, no item */
  public static final int PURE_TRAIN = 1;

  /*------------------------------------------------------------------*/
  /*  instance variables                                              */
  /*------------------------------------------------------------------*/
  /** the underlying item base */
  protected IdMap    ibase;
  /** the associated numeric identifier (item name is numeric) */
  protected int      nid;
  /** the associated item identifier */
  protected int      item;
  /** the times of the train */
  protected double[] times;
  /** the current number of times */
  protected int      cnt;
  /** the minimal time */
  protected double   min;
  /** the maximal time */
  protected double   max;

  /*------------------------------------------------------------------*/
  /** Create an empty (spike/event) train.
   *  @param  item  the item for which to create the train
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Train (String item)
  { this(null, item); }

  /*------------------------------------------------------------------*/
  /** Create an empty (spike/event) train.
   *  @param  ibase the underlying item base
   *  @param  item  the item for which to create the train
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Train (IdMap ibase, String item)
  {                             /* --- create a train */
    this.ibase = (ibase != null) ? ibase : new IdMap();
    if (item == null)           /* if no item given, use an index */
      item = String.valueOf(this.nid = this.ibase.size()+1);
    else {                      /* otherwise check for numeric item */
      try { this.nid = Integer.parseInt(item); }
      catch (NumberFormatException e) { this.nid = Integer.MIN_VALUE; }
    }                           /* note a numeric item identifier */
    this.item = this.ibase.add(item);
    this.cnt  = 0;              /* note the item and init. fields */
    this.min  = Double.POSITIVE_INFINITY;
    this.max  = Double.NEGATIVE_INFINITY;
  }  /* Train() */

  /*------------------------------------------------------------------*/
  /** Get the associated item (i.e., its identifier).
   *  @return the identifier of the associated item
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int getItem ()
  { return this.item; }

  /*------------------------------------------------------------------*/
  /** Get the associated item (i.e., its identifier).
   *  @return the identifier of the associated item
   *  @since  2016.08.25 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int getItemId ()
  { return this.item; }

  /*------------------------------------------------------------------*/
  /** Get the name of the associated item.
   *  @return the name of the associated item
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public String getItemName ()
  { return (String)this.ibase.get(this.item); }

  /*------------------------------------------------------------------*/
  /** Get the length/size of the (spike/event) train.
   *  @return the length/size of the (spike/event) train
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int getLength ()
  { return this.cnt; }

  /*------------------------------------------------------------------*/
  /** Get the length/size of the (spike/event) train.
   *  @return the length/size of the (spike/event) train
   *  @since  2016.08.25 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int getSize ()
  { return this.cnt; }

  /*------------------------------------------------------------------*/
  /** Get the minimum (spike/event) time.
   *  @return the minimum time
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double getMinTime ()
  { return this.min; }

  /*------------------------------------------------------------------*/
  /** Get the maximum (spike/event) time.
   *  @return the maximum time
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double getMaxTime ()
  { return this.max; }

  /*------------------------------------------------------------------*/
  /** Get a (spike/event) time of the train.
   *  @param  i the index of the time
   *  @return the value of the ith time
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double getTime (int i)
  { return this.times[i]; }

  /*------------------------------------------------------------------*/
  /** Get all (spike/event) times of the train
   * (array length/size fits only after packing).
   *  @return the array of times of the (spike/event) train
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double[] getAllTimes ()
  { return this.times; }

  /*------------------------------------------------------------------*/
  /** Add a (spike/event) time to the train.
   *  @param  time  the (spike/event) time to add
   *  @return the index of the new time
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int addTime (double time)
  {                             /* --- add a time to a train */
    double[] p;                 /* buffer for reallocation */
    int      size;              /* size of the times array */

    size = (this.times != null) ? this.times.length : 0;
    if (this.cnt >= size) {     /* if the times array is full */
      size += (size > BLKSIZE) ? size >> 1 : BLKSIZE;
      p = new double[size];     /* create an (enlarged) times array */
      if (this.times != null)   /* if there are existing times */
        System.arraycopy(this.times, 0, p, 0, this.cnt);
      this.times = p;           /* copy the existing items */
    }                           /* and set the (new) item array */
    this.times[this.cnt] = time;/* store the new time */
    if (time < this.min) this.min = time;  /* update the */
    if (time > this.max) this.max = time;  /* time range */
    return this.cnt++;          /* return the time index */
  }  /* addTime() */

  /*------------------------------------------------------------------*/
  /** Sort a (spike/event) train (that is, the times in it).
   *  @since  2013.12.03 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void sort ()
  { Arrays.sort(this.times, 0, this.cnt); }

  /*------------------------------------------------------------------*/
  /** Filter a (spike/event) train.
   *  @param  start the start of the filter interval
   *  @param  end   the end   of the filter interval
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void filter (double start, double end)
  {                             /* --- filter a train */
    int    i, k;                /* loop variables */
    double p;                   /* buffer for a time */

    if ((start <= Double.NEGATIVE_INFINITY)
    &&  (end   >= Double.POSITIVE_INFINITY)) return;
    this.min = Double.POSITIVE_INFINITY;
    this.max = Double.NEGATIVE_INFINITY;
    for (i = k = 0; i < this.cnt; i++) {
      if ((this.times[i] < start) || (this.times[i] > end))
        continue;               /* skip times outside given interval */
      this.times[k++] = p = this.times[i];
      if (p < this.min) this.min = p;
      if (p > this.max) this.max = p;
    }                           /* update the time range */
    this.cnt = k;               /* set the new number of times */
    this.pack();                /* and pack the filtered train */
  }  /* filter() */

  /*------------------------------------------------------------------*/
  /** Pack a (spike/event) train, i.e., optimize memory usage.
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void pack ()
  {                             /* --- pack a train */
    double[] p;                 /* buffer for reallocation */
    int      size;              /* size of the times array */

    size = (this.times != null) ? this.times.length : 0;
    if (size <= this.cnt) return;
    p = new double[this.cnt];   /* shrink the times array */
    System.arraycopy(this.times, 0, p, 0, this.cnt);
    this.times = p;             /* set the shrunk times array */
  }  /* pack() */

  /*------------------------------------------------------------------*/
  /** Write a (spike/event) train.
   *  @param  writer the table writer to write to
   *  @param  mode   the write mode
   *          (<code>ITEM_TRAIN</code> or <code>PURE_TRAIN</code>)
   *  @throws IOException if a write error occurs
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void write (TableWriter writer, int mode) throws IOException
  {                             /* --- write a train set */
    if (mode == ITEM_TRAIN)     /* write the train item */
      writer.write(this.getItemName());
    for (int i = 0; i < this.cnt; i++) /* write the times */
      writer.write(String.valueOf(this.times[i]), i >= this.cnt-1);
  }  /* write() */

  /*------------------------------------------------------------------*/
  /** Create a 'missing item' I/O exception.
   *  @param  scan the scanner to read from
   *  @return a 'missing item' I/O exception.
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private static IOException misItemEx (TableReader reader)
  { return new IOException("missing item " +reader.rno()); }

  /*------------------------------------------------------------------*/
  /** Create a 'malformed number' I/O exception.
   *  @param  scan the scanner to read from
   *  @return a 'malformed number' I/O exception.
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private static IOException badNumEx (TableReader reader)
  { return new IOException("malformed number '" +reader.field +"'"
                           +reader.rno()); }

  /*------------------------------------------------------------------*/
  /** Parse a (spike/event) train.
   *  @param  ibase   the underlying item base
   *  @param  reader  the reader to read from
   *  @param  mode    the record format mode
   *  @return the parsed (spike/event) train
   *  @throws IOException if a read error occurs
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static Train parse(IdMap ibase, TableReader reader, int mode)
    throws IOException
  {                             /* --- parse a train */
    Train  train;               /* parsed train */
    String name;                /* name of associated item */
    double p;                   /* next time of the train */

    if (reader.nextField() < 0) /* if at the end of the input, */
      return null;              /* there is no train to read */
    /* Note that nextField() can return a negative value only   */
    /* on the first call, since a value of 0 (field separator)  */
    /* cannot be followed immediately by a negative value and   */
    /* a positive value (record separator) terminates the loop. */
    if      (mode != ITEM_TRAIN) name = String.valueOf(ibase.size()+1);
    else if (reader.field != null) name = reader.field;
    else throw Train.misItemEx(reader);
    train = new Train(ibase, name); /* create an empty train */
    if (mode == ITEM_TRAIN) {   /* if to parse a train with an item */
      if (reader.delim > 0) return train;
      reader.nextField();       /* check for end of record */
    }                           /* and read the next field */
    while (true) {              /* read the times of the train */
      if ((reader.field == null) && (reader.delim > 0)) break;
      try { p = Double.parseDouble(reader.field); }
      catch (NumberFormatException e) { throw Train.badNumEx(reader); }
      train.addTime(p);         /* add the time to the train */
      if (reader.delim != 0) break;
      reader.nextField();       /* check for the end of the record, */
    }                           /* otherwise read the next field */
    train.pack();               /* pack the parsed train */
    return train;               /* and then return it */
  }  /* parse() */

  /*------------------------------------------------------------------*/
  /** Parse a (spike/event) train.
   *  @param  ibase  the underlying item base
   *  @param  reader the reader to read from
   *  @param  mode   the record format mode
   *  @return the parsed (spike/event) train
   *  @throws IOException if a read error occurs
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static Train parse (IdMap ibase, Reader reader, int mode)
    throws IOException
  { return Train.parse(ibase, new TableReader(reader), mode); }

  /*------------------------------------------------------------------*/
  /** Parse a (spike/event) train.
   *  @param  ibase the underlying item base
   *  @param  desc  the string description to parse
   *  @param  mode  the record format mode
   *  @return the parsed (spike/event) train
   *  @throws IOException if a read error occurs
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static Train parse (IdMap ibase, String desc, int mode)
    throws IOException
  { return Train.parse(ibase, new TableReader(desc), mode); }

  /*------------------------------------------------------------------*/
  /** Parse a (spike/event) train.
   *  @param  ibase the underlying item base
   *  @param  inp   the input stream to read from
   *  @param  mode  the record format mode
   *  @return the parsed (spike/event) train
   *  @throws IOException if a read error occurs
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static Train parse (IdMap ibase, InputStream inp,int mode)
    throws IOException
  { return Train.parse(ibase, new TableReader(inp), mode); }

  /*------------------------------------------------------------------*/
  /** Create a string description of a (spike/event) train.
   *  @return the created string description
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public String toString ()
  {                             /* --- create a string representation */
    int           i;            /* loop variable */
    StringBuilder b;            /* created string representation */

    b = new StringBuilder(this.getItemName());
    for (i = 0; i < this.cnt; i++) { /* list the times of the train */
      b.append(" "); b.append(String.valueOf(this.times[i])); }
    return b.toString();        /* return the string representation */
  }  /* toString() */

}  /* class Train */
