/*----------------------------------------------------------------------
  File    : TrainSet.java
  Contents: train set management for visualization
  Author  : Christian Borgelt
  History : 2013.11.27 file created
            2013.11.29 train set parsing functions added/completed
            2013.12.13 sorting the trains by their item added
            2014.10.03 function getAllTrains() added
            2015.08.17 generally renamed point -> time
            2016.08.26 functions getItem[|Id|Name] added/fixed
----------------------------------------------------------------------*/
package coconad;

import java.util.Arrays;
import java.util.Comparator;
import java.io.IOException;
import java.io.InputStream;
import java.io.FileReader;
import java.io.Reader;
import java.io.Serializable;

import util.IdMap;
import util.TableReader;
import util.TableWriter;

/*--------------------------------------------------------------------*/
/** Class for a set of trains.
 *  @author Christian Borgelt
 *  @since  2013.11.27 */
/*--------------------------------------------------------------------*/
public class TrainSet implements Serializable {

  /*------------------------------------------------------------------*/
  /*  constants                                                       */
  /*------------------------------------------------------------------*/
  private static final long serialVersionUID = 0x00010000L;
  /** the block size for the train array */
  private static final int BLKSIZE   = 32;
  /** read/write mode: train with preceding item */
  public static final int ITEM_TRAIN = Train.ITEM_TRAIN;
  /** read/write mode: pure train, that is, no item */
  public static final int PURE_TRAIN = Train.PURE_TRAIN;
  /** read/write mode: pair of item and time */
  public static final int ITEM_TIME  = 2;
  /** read/write mode: pair of time and item */
  public static final int TIME_ITEM  = 3;

  /*------------------------------------------------------------------*/
  /*  instance variables                                              */
  /*------------------------------------------------------------------*/
  /** the underlying item base */
  protected IdMap   ibase;
  /** the set of trains */
  protected Train[] trains;
  /** the current number of trains */
  protected int     cnt;
  /** the maximal length of a train */
  protected int     maxlen;
  /** the minimum event/spike time */
  protected double  min;
  /** the maximum event/spike time */
  protected double  max;

  /*------------------------------------------------------------------*/
  /** Create an empty set of trains.
   *  @since  2013.11.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public TrainSet ()
  { this(null); }

  /*------------------------------------------------------------------*/
  /** Create an empty set of trains.
   *  @param  ibase the underlying item base
   *  @since  2013.11.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public TrainSet (IdMap ibase)
  {                             /* --- create a train set */
    this.ibase  = (ibase != null) ? ibase : new IdMap();
    this.maxlen = this.cnt = 0; /* note item base and init. fields */
    this.min    = Double.POSITIVE_INFINITY;
    this.max    = Double.NEGATIVE_INFINITY;
  }  /* TrainSet() */

  /*------------------------------------------------------------------*/
  /** Get the underlying item base.
   *  @return the underlying item base
   *  @since  2013.11.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public IdMap getItemBase ()
  { return this.ibase; }

  /*------------------------------------------------------------------*/
  /** Get the item identifier of a (spike/event) train.
   *  @param  i  the index of the train of which to get the item
   *  @return the item of the train with index <code>i</code>
   *  @since  2016.08.26 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int getItem (int i)
  { return this.trains[i].getItem(); }

  /*------------------------------------------------------------------*/
  /** Get the item identifier of a (spike/event) train.
   *  @param  i  the index of the train of which to get the item
   *  @return the item of the train with index <code>i</code>
   *  @since  2016.08.26 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int getItemId (int i)
  { return this.trains[i].getItemId(); }

  /*------------------------------------------------------------------*/
  /** Get the item name of a (spike/event) train.
   *  @param  i  the index of the train of which to get the item name
   *  @return the item name of the train with index <code>i</code>
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public String getItemName (int i)
  { return this.trains[i].getItemName(); }

  /*------------------------------------------------------------------*/
  /** Get the number of (spike/event) trains.
   *  @return the number of trains
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int getSize ()
  { return this.cnt; }

  /*------------------------------------------------------------------*/
  /** Get a train.
   *  @param  i the index of the train
   *  @return the train with index <code>i</code>
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Train getTrain (int i)
  { return this.trains[i]; }

  /*------------------------------------------------------------------*/
  /** Get all trains as arrays of floating point values.
   *  @return all trains as arrays of floating point values
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double[][] getAllTrains ()
  {                             /* --- get all trains */
    this.pack();                /* ensure proper train lengths */
    double[][] t = new double[this.cnt][];
    for (int i = 0; i < this.cnt; i++)
      t[i] = this.trains[i].getAllTimes();
    return t;                   /* collect the trains and */
  }  /* getAllTrains() */       /* return the resulting array */

  /*------------------------------------------------------------------*/
  /** Add a train.
   *  @param  train the train to add
   *  @return the index of the train in the set
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int addTrain (Train train)
  {                             /* --- add a train */
    Train[] t;                  /* buffer for reallocation */
    int     size;               /* new array size */
    double  p;                  /* buffer for a time */

    size = (this.trains != null) ? this.trains.length : 0;
    if (this.cnt >= size) {     /* if the train array is full */
      size += (size > BLKSIZE) ? size >> 1 : BLKSIZE;
      t = new Train[size];      /* create an (enlarged) train array */
      if (this.trains != null)  /* copy already existing times */
        System.arraycopy(this.trains, 0, t, 0, this.cnt);
      this.trains = t;          /* set the (new) train array */
    }
    this.trains[this.cnt] = train;   /* add the new train */
    size = train.getLength();   /* adapt the maximal train length */
    if (size > this.max) this.max = size;
    p = train.getMinTime();     /* update the time range */
    if (p < this.min) this.min = p;
    p = train.getMaxTime();
    if (p > this.max) this.max = p;
    return this.cnt++;          /* return the train identifier */
  }  /* addTrain() */

  /*------------------------------------------------------------------*/
  /** Get the maximal length of a train.
   *  @return the maximal size of a train
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int getMaxLength ()
  { return this.maxlen; }

  /*------------------------------------------------------------------*/
  /** Sort trains by their item names.
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void sort ()
  {                             /* --- sort the trains by items */
    int   i;                    /* loop variable */
    int[] map;                  /* map from old to new items */

    for (i = this.cnt; --i >= 0; )
      if (this.trains[i].nid <= Integer.MIN_VALUE) break;
    if (i <= 0) {               /* if all items are numeric */
      Arrays.sort(this.trains, 0, this.cnt,
        new Comparator<Train> () {
          public int compare (Train t1, Train t2) {
            if     (t1.nid == t2.nid) return 0;
            return (t1.nid <  t2.nid) ? -1 : +1; }
          public boolean equals (Train t1, Train t2) {
            return (t1.nid == t2.nid); }
      }); }                     /* compare numeric identifiers */
    else {                      /* if some items are actual strings */
      Arrays.sort(this.trains, 0, this.cnt,
        new Comparator<Train> () {
          public int compare (Train t1, Train t2) {
            return t1.getItemName().compareTo(t2.getItemName()); }
          public boolean equals (Train t1, Train t2) {
            return t1.getItemName().equals(t2.getItemName()); }
      });                       /* compare the item names */
    }                           /* (sort the trains by item) */
    map = new int[this.cnt];    /* create an item map and */
    for (i = 0; i < cnt; i++)   /* fill it from the item mappings */
      map[this.trains[i].item] = i;
    this.ibase.reorder(map);    /* reorder the item base */
  }  /* sort() */

  /*------------------------------------------------------------------*/
  /** Sort a train set (that is, sort all contained trains).
   *  @since  2013.12.03 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void sortTimes ()
  {                             /* --- sort times of a train set */
    for (int i = 0; i < this.cnt; i++)
      this.trains[i].sort();    /* traverse and sort all trains */
  }  /* sortTimes() */

  /*------------------------------------------------------------------*/
  /** Filter a train set.
   *  @param  start the start of the filter interval
   *  @param  end   the end   of the filter interval
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void filter (double start, double end)
  {                             /* --- filter a train set */
    int    i, n;                /* loop variable, buffer */
    double p;                   /* buffer for a time */

    if ((start <= Double.NEGATIVE_INFINITY)
    &&  (end   >= Double.POSITIVE_INFINITY)) return;
    this.maxlen = 0;            /* reinit. max. length and time range */
    this.min = Double.POSITIVE_INFINITY;
    this.max = Double.NEGATIVE_INFINITY;
    for (i = 0; i < this.cnt; i++) {
      this.trains[i].filter(start, end);
      n = this.trains[i].getLength();
      if (n > this.maxlen) this.maxlen= n;
      p = this.trains[i].getMinTime();
      if (p < this.min)    this.min = p;
      p = this.trains[i].getMaxTime();
      if (p > this.max)    this.max = p;
    }                           /* update max. length and time range */
  }  /* filter() */

  /*------------------------------------------------------------------*/
  /** Pack the trains of a set, i.e., optimize memory usage.
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void pack ()
  {                             /* --- pack all trains */
    for (int i = 0; i < this.cnt; i++)
      this.trains[i].pack();    /* traverse and pack the trains */
  }  /* pack() */

  /*------------------------------------------------------------------*/
  /** Update the time range variables.
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void getRange()
  {                             /* --- update time range variables */
    int      i;                 /* item identifier */
    double   p;                 /* buffer for a time */

    this.min = Double.POSITIVE_INFINITY;
    this.max = Double.NEGATIVE_INFINITY;
    for (i = 0; i < this.cnt; i++) {
      p = this.trains[i].getMinTime();
      if (p < this.min) this.min = p;
      p = this.trains[i].getMaxTime();
      if (p > this.max) this.max = p;
    }                           /* update the time range */
  }  /* getRange() */

  /*------------------------------------------------------------------*/
  /** Get the minimum time.
   *  @return the minimum time
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double getMinTime ()
  { return this.min; }

  /*------------------------------------------------------------------*/
  /** Get the maximum time.
   *  @return the maximum time
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double getMaxTime ()
  { return this.max; }

  /*------------------------------------------------------------------*/
  /** Write a set of trains.
   *  @param  writer the table writer to write to
   *  @param  mode   the read mode of the writer
   *  @throws IOException if a write error occurs
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void write (TableWriter writer, int mode) throws IOException
  {                             /* --- write a train set */
    int    i, k;                /* loop variables */
    Train  t;                   /* to traverse the trains */
    String n;                   /* name of associated item */

    if      (mode == ITEM_TIME) {  /* if pairs of item and time */
      for (i = 0; i < this.cnt; i++) {
        t = this.trains[i];     /* traverse the trains */
        n = t.getItemName();    /* get the associated item */
        for (k = 0; k < t.cnt; k++) {
          writer.write(n);      /* traverse the times of the train */
          writer.writeLast(String.valueOf(t.times[k]));
        }                       /* write item name and time */
      } }
    else if (mode == TIME_ITEM) {  /* if pairs of time and item */
      for (i = 0; i < this.cnt; i++) {
        t = this.trains[i];     /* traverse the trains */
        n = t.getItemName();    /* get the associated item */
        for (k = 0; k < t.cnt; k++) {
          writer.write(String.valueOf(t.times[k]), k >= t.cnt-1);
          writer.writeLast(n);  /* traverse the times of the train */
        }                       /* and write time and item name */
      } }
    else {                      /* if one train per record */
      for (i = 0; i < this.cnt; i++)
        this.trains[i].write(writer, mode);
    }                           /* write the trains */
  }  /* write() */

  /*------------------------------------------------------------------*/
  /** Create a 'missing item' I/O exception.
   *  @param  scan the scanner to read from
   *  @return a 'missing item' I/O exception.
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private static IOException misItemEx (TableReader reader)
  { return new IOException("missing item " +reader.rno()); }

  /*------------------------------------------------------------------*/
  /** Create a 'missing time' I/O exception.
   *  @param  scan the scanner to read from
   *  @return a 'missing time' I/O exception.
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private static IOException misTimeEx (TableReader reader)
  { return new IOException("missing time " +reader.rno()); }

  /*------------------------------------------------------------------*/
  /** Create a 'too many fields' I/O exception.
   *  @param  scan the scanner to read from
   *  @return a 'too many fields' I/O exception.
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private static IOException fieldCntEx (TableReader reader)
  { return new IOException("too many fields " +reader.rno()); }

  /*------------------------------------------------------------------*/
  /** Create a 'malformed number' I/O exception.
   *  @param  scan the scanner to read from
   *  @return a 'malformed number' I/O exception.
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private static IOException badNumEx (TableReader reader)
  { return new IOException("malformed number '" +reader.field +"'"
                           +reader.rno()); }

  /*------------------------------------------------------------------*/
  /** Parse a set of (spike/event) trains.
   *  @param  ibase  the underlying item base
   *  @param  reader the table reader to read from
   *  @param  mode   the read mode of the writer
   *  @return the parsed set of (spike/event) trains
   *  @throws IOException if a read error occurs
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static TrainSet parse
    (IdMap ibase, TableReader reader, int mode) throws IOException
  {                             /* --- parse a train set */
    TrainSet trains;            /* created train set */
    Train    t;                 /* to traverse the trains */
    int      i;                 /* item identifier */
    double   p;                 /* buffer for a time */
    String   item;              /* the name of the current item */

    trains = new TrainSet(ibase);
    if ((mode == ITEM_TRAIN)    /* if the input format */
    ||  (mode == PURE_TRAIN)) { /* has one train per record */
      while (true) {            /* train read loop */
        t = Train.parse(trains.ibase, reader, mode);
        if (t == null) break;   /* read the next train */
        trains.addTrain(t);     /* and add it to the set */
      } }
    else {                      /* if the input consists of pairs */
      while (reader.nextField() >= 0) { /* while more fields */
        if ((reader.field == null) && (reader.delim > 0))
          continue;             /* skip empty records */
        if (mode == ITEM_TIME) {/* if pair of item and time */
          if (reader.field == null) throw TrainSet.misItemEx(reader);
          i = trains.ibase.add(item = reader.field);
          if (reader.delim != 0)    throw TrainSet.misTimeEx(reader);
          reader.nextField();   /* get item first and then time */
          if (reader.field == null) throw TrainSet.misTimeEx(reader);
          try { p = Double.parseDouble(reader.field); }
          catch (NumberFormatException e) {
            throw TrainSet.badNumEx(reader); }
          if (reader.delim <= 0)    throw TrainSet.fieldCntEx(reader); }
        else {                  /* if pair of time and item */
          if (reader.field == null) throw TrainSet.misTimeEx(reader);
          try { p = Double.parseDouble(reader.field); }
          catch (NumberFormatException e) {
            throw TrainSet.badNumEx(reader); }
          if (reader.delim != 0)    throw TrainSet.misItemEx(reader);
          reader.nextField();   /* get item first and then time */
          if (reader.field == null) throw TrainSet.misItemEx(reader);
          i = trains.ibase.add(item = reader.field);
          if (reader.delim <= 0)    throw TrainSet.fieldCntEx(reader);
        }
        if (i >= trains.cnt)    /* add a new train to the set */
          trains.addTrain(new Train(trains.ibase, item));
        trains.trains[i].addTime(p);
      }                         /* add the time read */
      trains.pack();            /* pack all trains */
      trains.getRange();        /* determine the time range */
      trains.sort();            /* sort trains by their item */
    }
    return trains;              /* return the created set */
  }  /* parse() */

  /*------------------------------------------------------------------*/
  /** Parse a set of (spike/event) trains.
   *  @param  ibase  the underlying item base
   *  @param  reader the reader to read from
   *  @param  mode   the read mode of the writer
   *  @return the parsed set of (spike/event) trains
   *  @throws IOException if a read error occurs
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static TrainSet parse (IdMap ibase, Reader reader, int mode)
    throws IOException
  { return TrainSet.parse(ibase, new TableReader(reader), mode); }

  /*------------------------------------------------------------------*/
  /** Parse a set of (spike/event) trains.
   *  @param  ibase the underlying item base
   *  @param  desc  the string description to parse
   *  @param  mode   the read mode of the writer
   *  @return the parsed set of (spike/event) trains
   *  @throws IOException if a read error occurs
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static TrainSet parse (IdMap ibase, String desc, int mode)
    throws IOException
  { return TrainSet.parse(ibase, new TableReader(desc), mode); }

  /*------------------------------------------------------------------*/
  /** Parse a set of (spike/event) trains.
   *  @param  ibase the underlying item base
   *  @param  inp   the input stream to read from
   *  @param  mode   the read mode of the writer
   *  @return the parsed set of (spike/event) trains
   *  @throws IOException if a read error occurs
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static TrainSet parse (IdMap ibase, InputStream inp, int mode)
    throws IOException
  { return TrainSet.parse(ibase, new TableReader(inp), mode); }

  /*------------------------------------------------------------------*/
  /** Main function for testing some basic functionality.
   *  <p>It is tried to parse the file that is given as the first
   *  command line argument as a set of trains.</p>
   *  @param  args the command line arguments
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static void main (String[] args)
  {                             /* --- main function for testing */
    TableReader reader;         /* table reader to read from */
    TableWriter writer;         /* table writer to write to */
    TrainSet    trains;         /* created trains set */
    int         minp = PURE_TRAIN; /* read  mode/input  format */
    int         mout = PURE_TRAIN; /* write mode/output format */

    try {                       /* try to read the given file */
      if (args.length <= 0)     /* if no arguments given */
        reader = new TableReader("0.0 0.1 0.2\n1.0 1.1 1.2 1.3\n"
                                +"2.0 2.1\n3.0 3.1 3.2\n\n");
      else {                    /* if a file argument is given */
        reader = new TableReader(new FileReader(args[0]));
        if ((args.length > 1) && (args[1].length() > 0)) {
          switch (args[1].charAt(0)) {
            case 'i': minp = ITEM_TIME;  break;
            case 'r': minp = TIME_ITEM;  break;
            case 't': minp = ITEM_TRAIN; break;
            case 'p': minp = PURE_TRAIN; break;
          }                     /* if there is a second argument, */
        }                       /* get the read mode */
        if ((args.length > 2) && (args[2].length() > 0)) {
          switch (args[2].charAt(0)) {
            case 'i': mout = ITEM_TIME;  break;
            case 'r': mout = TIME_ITEM;  break;
            case 't': mout = ITEM_TRAIN; break;
            case 'p': mout = PURE_TRAIN; break;
          }                     /* if there is a second argument, */
        }                       /* get the write mode */
      }
      trains = TrainSet.parse(null, reader, minp);
      reader.close();           /* parse the train set */
      writer = new TableWriter(System.out);
      trains.write(writer, mout);
      writer.close(); }         /* write the train set */
    catch (IOException e) {     /* catch and report i/o errors */
      System.err.println(e.getMessage()); return; }
  }  /* main() */

}  /* class TrainSet */
